############################################################################
#                                                                          #
#   Copyright (C) 2011 Po-Hsien Lee <p.lee@mx.uni-saarland.de>             #
#   All rights reserved.                                                   #
#                                                                          #
#   This file is part of PROPORES.                                         #
#                                                                          #
#   PROPORES is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation, either version 3 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   PROPORES is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License      #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
#                                                                          #
############################################################################
#                                                                          #
#   Pore_ID.pl                                                             #
#                                                                          #
#   Programmed by Po-Hsien Lee <p.lee@mx.uni-saarland.de>, 2011            #
#   Center for Bioinformatics, Saarland University, Germany                #
#                                                                          #
#   Pore_ID.pl is a program for identifying pores of protein structures.   #
#                                                                          #
############################################################################

#!/usr/bin/perl

### include necessary perl module and initialization
use Parameter;
use PDBreader;
use Vector;
use POSIX;
use KD_tree_NN;
use Getopt::Std;

$pdb_obj=PDBreader->new();
$vector_op=Vector->new();
$kd_tree_op=KD_tree_NN->new();
($Tuser,$Tsystem,$Tchild,$Tchildsystem)=times;


### get and check input parameters
($pdb_file, $grid_side_length, $probe_radius, $shallow_criteria, $output_prefix)=&get_check_options();


### some derived constants
$grid_radius=$grid_side_length/2.0;
$g_s_l_cubic=$grid_side_length**3;
$solvent_radius=$probe_radius; 
$max_vdw_r=$Parameter::vdw_radius{'S'};
$solvent_margin=ceil($solvent_radius*2/$grid_side_length);
$solvent_margin=(($solvent_margin%2)==0)?($solvent_margin+1):$solvent_margin;
$marginal_level=$solvent_margin+ceil($max_vdw_r/$grid_side_length);


### useful pre-defined array 
@search_seq_1=();
&search_sequence(1, \@search_seq_1);


### Step 1 : Read in pdb file 
@ATOM=();
$pdb_obj->pdbparser($pdb_file, \@ATOM, "deHETATM");
print "Step 1 : Read input pdb file : finished\n";



### Step 2 : Definition range of the protein in the grid
$x_upper=$y_upper=$z_upper=-99999;
$x_lower=$y_lower=$z_lower=99999;
for($i=0;$i<=$#ATOM;$i++)
{
        &pdb_boundary(\$x_upper,\$x_lower,$ATOM[$i][9]);
        &pdb_boundary(\$y_upper,\$y_lower,$ATOM[$i][10]);
        &pdb_boundary(\$z_upper,\$z_lower,$ATOM[$i][11]);
}

&grid_boundary(\$x_upper,\$x_lower,$grid_side_length);
&grid_boundary(\$y_upper,\$y_lower,$grid_side_length);
&grid_boundary(\$z_upper,\$z_lower,$grid_side_length);

$pro_starti=$marginal_level;
$sol_starti=floor($solvent_margin/2);
$pro_x_endi=&round(($x_upper-$x_lower)/$grid_side_length)+$marginal_level;
$pro_y_endi=&round(($y_upper-$y_lower)/$grid_side_length)+$marginal_level;
$pro_z_endi=&round(($z_upper-$z_lower)/$grid_side_length)+$marginal_level;

$sol_x_endi=$pro_x_endi+$marginal_level-$sol_starti;
$sol_y_endi=$pro_y_endi+$marginal_level-$sol_starti;
$sol_z_endi=$pro_z_endi+$marginal_level-$sol_starti;
print "Step 2 : Definition range for loop calculation : finished\n";



### Step 3 : gridization of atoms and vdW extension 
($user,$system,$child,$childsystem)=times;
@vdw_boundary=();
for($i=0;$i<=$#ATOM;$i++)
{
	($ATOM[$i][14],$ATOM[$i][15],$ATOM[$i][16])=&gridize($ATOM[$i][9],$ATOM[$i][10],$ATOM[$i][11]);
	$ATOM[$i][17]=$i;  # $ATOM[][17] is specific for KD_tree 
	&vdw_extension($i,\@ATOM,\@vdw_boundary);
}
($user1,$system1,$child1,$childsystem1)=times;
$user1-=$user;
$system1-=$system;
$child1-=$child;
$childsystem1-=$childsystem;
print "\nStep 3 : VDW extension : finished\n";
printf "running time = %.1f (s)\n", $user1;



### Step 4 : Scan along XYZ axes
($user,$system,$child,$childsystem)=times;
@face_number=(); @mouth_open=();
&perform_face_count('x',\@vdw_boundary,\@face_number,\@mouth_open);
&perform_face_count('y',\@vdw_boundary,\@face_number,\@mouth_open);
&perform_face_count('z',\@vdw_boundary,\@face_number,\@mouth_open);
($user1,$system1,$child1,$childsystem1)=times;
$user1-=$user;
$system1-=$system;
$child1-=$child;
$childsystem1-=$childsystem;
print "\nStep 4 : Scan along XYZ axes : finished\n";
printf "running time = %.1f (s)\n", $user1;



### Step 5 : Orientational complementation (cylindrical scheme)
($user,$system,$child,$childsystem)=times;
&orientational_complement(\@ATOM,\@vdw_boundary,\@face_number);
($user1,$system1,$child1,$childsystem1)=times;
$user1-=$user;
$system1-=$system;
$child1-=$child;
$childsystem1-=$childsystem;
print "\nStep 5 : orientational complement : finished\n";
printf "running time = %.1f (s)\n", $user1;



### Step 6 : Seal the gaps inside the protein
($user,$system,$child,$childsystem)=times;
@not_covered_raw=(); 
for($i=$sol_starti-1; $i<=$sol_x_endi+1; $i++)
{
	for($j=$sol_starti-1; $j<=$sol_y_endi+1; $j++)
	{
		for($k=$sol_starti-1; $k<=$sol_z_endi+1; $k++)
		{
			if($face_number[$i][$j][$k]<40 && $vdw_boundary[$i][$j][$k][0]!=1)
			{
				$not_covered_raw[$i][$j][$k]=1;
			}
		}
	}
}
@background=(); $background[$sol_starti][$sol_starti][$sol_starti]=1;
&union($sol_starti, $sol_starti, $sol_starti, \@background, \@not_covered_raw, \@search_seq_1);

#open FH, "> lost.pdb " or die "$!"; $temp=0;
for($i=$sol_starti; $i<=$sol_x_endi; $i++)
{
	for($j=$sol_starti; $j<=$sol_y_endi; $j++)
	{
		for($k=$sol_starti; $k<=$sol_z_endi; $k++)
		{
			if($not_covered_raw[$i][$j][$k]==1 && $background[$i][$j][$k]!=1 )
			{
				#$temp++;
				#($coor_i,$coor_j,$coor_k)=&inverse($i, $j, $k);
				#&print_pseduPDB(FH,$temp,1,$coor_i,$coor_j,$coor_k);
				$face_number[$i][$j][$k]=40;
			}
		}
	}
}
#close FH;
@not_covered_raw=();

($user1,$system1,$child1,$childsystem1)=times;
$user1-=$user;
$system1-=$system;
$child1-=$child;
$childsystem1-=$childsystem;
print "\nStep 6 : Seal the gaps inside the protein : finished\n";
printf "running time = %.1f (s)\n", $user1;



### Step 7 : fit largest ball into all grid voxel by KD-tree algorithm
($user,$system,$child,$childsystem)=times;
@largest_ball=();
$tree={};
$tree=$kd_tree_op->construct_kd_tree(\@ATOM,0);
for($i=$sol_starti; $i<=$sol_x_endi; $i++)
{
	for($j=$sol_starti; $j<=$sol_y_endi; $j++)
	{
		for($k=$sol_starti; $k<=$sol_z_endi; $k++)
		{
			if( $vdw_boundary[$i][$j][$k][0] != 1 && $face_number[$i][$j][$k]>=40)
			{
				@coor=();
				($coor[0],$coor[1],$coor[2])=&inverse($i,$j,$k);
				($atm_idx,$min_r)=$kd_tree_op->nnsearch_kd_tree_r1($tree,0,\@ATOM,@coor);
				$kd_tree_op->nnsearch_kd_tree_r2($tree,0,\$atm_idx,\$min_r,\@ATOM,@coor);
				$largest_ball[$i][$j][$k][0]=$min_r;
				$largest_ball[$i][$j][$k][1]=$atm_idx;
			}
		}
	}
}
($user1,$system1,$child1,$childsystem1)=times;
$user1-=$user;
$system1-=$system;
$child1-=$child;
$childsystem1-=$childsystem;
print "\nStep 7 : fit in largest ball : finished\n";
printf "running time = %.1f (s)\n", $user1;



### Step 8 : Remove shallow pore region on the protein surface
($user,$system,$child,$childsystem)=times;
$sov_d_sq=($shallow_criteria/$grid_side_length)**2;
$level_shallow=ceil($shallow_criteria/$grid_side_length);
@search_seq_shallow=();
&search_sequence($level_shallow, \@search_seq_shallow);
for($i=$sol_starti; $i<=$sol_x_endi; $i++)
{
	for($j=$sol_starti; $j<=$sol_y_endi; $j++)
	{
		for($k=$sol_starti; $k<=$sol_z_endi; $k++)
		{
			$sum=0;
			for($m=1; $m<=6; $m++)
			{
				$p=$i+$search_seq_1[$m][1];
				$q=$j+$search_seq_1[$m][2];
				$r=$k+$search_seq_1[$m][3];
				$sum+=$background[$p][$q][$r];
			}

			if( $sum < 6 && $sum > 0 )
			{
				for($m=1; $m<=$#search_seq_shallow; $m++)
				{
					$D_sq=$search_seq_shallow[$m][0];
					if( $D_sq <= $sov_d_sq )
					{
						$p=$i+$search_seq_shallow[$m][1];
						$q=$j+$search_seq_shallow[$m][2];
						$r=$k+$search_seq_shallow[$m][3];
						if( $vdw_boundary[$p][$q][$r][0]!=1 && $face_number[$p][$q][$r]>=40 )
						{
							$face_number[$p][$q][$r]=0;
						}
					}
				}
			}
		}
	}
}
($user1,$system1,$child1,$childsystem1)=times;
$user1-=$user;
$system1-=$system;
$child1-=$child;
$childsystem1-=$childsystem;
print "\nStep 8 : Remove shallow pore region on the surface : finished\n";
printf "running time = %.1f (s)\n", $user1;



### Step 9 : Cluster pore grid voxels and identify pore volumn and pore-lining residues
($user,$system,$child,$childsystem)=times;
@void_nucleus=();
for($i=$sol_starti; $i<=$sol_x_endi; $i++)
{
	for($j=$sol_starti; $j<=$sol_y_endi; $j++)
	{
		for($k=$sol_starti; $k<=$sol_z_endi; $k++)
		{
			if($face_number[$i][$j][$k]>=40 && $largest_ball[$i][$j][$k][0]>=$solvent_radius)
			{
				$void_nucleus[$i][$j][$k]=1;
			}
		}
	}
}

@pore_volume=(); @pore_cluster=(); 
&cluster(\@void_nucleus,\@pore_cluster,\@pore_volume,\@search_seq_1);

@search_seq_probe=();
$level=ceil(($solvent_radius-$grid_radius)/$grid_side_length);
&search_sequence($level, \@search_seq_probe);
&complete_volume(\@pore_volume,\@void_nucleus,\@face_number,\@pore_cluster, \@search_seq_probe, \@search_seq_1);
@pore_volume=sort { $a->[1] <=> $b->[1] }(@pore_volume);
@pore_volume=reverse(@pore_volume);

@COM=();
($com_x, $com_y, $com_z, $total_mass)=$pdb_obj->pdb_com(\@ATOM,"heavy");
($COM[0], $COM[1], $COM[2])=&gridize($com_x, $com_y, $com_z);
&pore_surface(\@pore_volume, \@pore_cluster, \@face_number, \@vdw_boundary, \@COM);

for( $i=0; $i<=$#pore_volume; $i++)
#for( $i=0; $i<=20; $i++)
{
	%lining_residue=();
	&extract_lining_residue(\@pore_volume,$i,\@vdw_boundary,\%lining_residue);
	$pore_volume_cubic=$pore_volume[$i][1]*$g_s_l_cubic;
	if($pore_volume_cubic < 11.5)  # 11.5 angstrom cubic is the volume of sphere with r = 1.4
	{
		last;
	}

	if($output_prefix ne "")
	{
		$prefix=$output_prefix."_pore";
	}
	else
	{
		$prefix="pore";
	}

	open FH, "> $prefix\_$i.pdb" or die "$!"; $temp=0;
	print FH "REMARK\tPore Volume = $pore_volume_cubic (angstrom cubic)\n";
	open FH2, ">$prefix\_$i.PTin" or die "$!";
	print FH2 "# Grid side length = $grid_side_length (angstrom)\n";
	print FH2 "# Pore volume = $pore_volume_cubic (angstrom cubic)\n";
	$length=$#{$pore_volume[$i][2]};
	for( $j=0; $j<=$length; $j++ )
	{
		$temp++;
		($coor_i,$coor_j,$coor_k)=&inverse(@{$pore_volume[$i][2][$j]});
		&print_pseduPDB(FH,$temp,1,$coor_i,$coor_j,$coor_k);
		printf FH2 "%.3f\t%.3f\t%.3f\t%f\t%d\t%.3f\n",$coor_i,$coor_j,$coor_k,$largest_ball[$pore_volume[$i][2][$j][0]][$pore_volume[$i][2][$j][1]][$pore_volume[$i][2][$j][2]][0], $pore_volume[$i][2][$j][3], $pore_volume[$i][2][$j][4];
	}
	close FH; close FH2;

	$pymol_text="";
	open FH,">$prefix\_$i.list" or die "$!";
	while($key1=each(%lining_residue))
	{
		while(($key2,$value)=each(%{$lining_residue{$key1}}))
		{
			print FH "$key1\t$key2\t$value\n";
			$pymol_text.="$key2+"
		}
	}
	print FH "$pymol_text\n";
	close FH;
}
($user1,$system1,$child1,$childsystem1)=times;
$user1-=$user;
$system1-=$system;
$child1-=$child;
$childsystem1-=$childsystem;
print "\nStep 9 : Pore volumn and lining residues finished\n";
printf "running time = %.1f (s)\n", $user1;



### Total running time ###
($Tuser1,$Tsystem1,$Tchild1,$Tchildsystem1)=times;
$Tuser1-=$Tuser;
$Tsystem1-=$Tsystem;
$Tchild1-=$Tchild;
$Tchildsystem1-=$Tchildsystem;
printf "\nJob Completed!\nTotal running time = %.1f (s)\n\n",$Tuser1;
exit;



### the rests are subroutins

### define the boundary of the protein in 3 dimensions
sub pdb_boundary
{
	my ($upper_ref, $lower_ref, $target)=@_;

	if($target > $$upper_ref)
	{
		$$upper_ref=$target;
	}
	if($target < $$lower_ref)
	{
		$$lower_ref=$target;
	}
}

### define the boundary of the protein in the grid
sub grid_boundary
{
	my ($upper_ref, $lower_ref, $grid_size)=@_;
	my $addon=0;

	$addon=0.5*(ceil(($$upper_ref-$$lower_ref)/$grid_size)*$grid_size-($$upper_ref-$$lower_ref));
	$$upper_ref+=$addon; $$lower_ref-=$addon;
}


### round value 
sub round
{
	my $value;
	$value=int(abs($_[0]) + 0.5) * ($_[0] < 0 ? -1 : 1);
	return $value;
}


### gridization of a point
### &gridize($x,$y,$z);
sub gridize
{
	my $grid_x=floor(($_[0]-$x_lower)/$grid_side_length)+$marginal_level;
	my $grid_y=floor(($_[1]-$y_lower)/$grid_side_length)+$marginal_level;
	my $grid_z=floor(($_[2]-$z_lower)/$grid_side_length)+$marginal_level;
	return ($grid_x,$grid_y,$grid_z);
}


### inversion of gridization
sub inverse
{
	my ($coor_i, $coor_j, $coor_k)=(0.0, 0.0, 0.0);

	$coor_i=($_[0]-$marginal_level)*$grid_side_length+$x_lower+$grid_radius;
	$coor_j=($_[1]-$marginal_level)*$grid_side_length+$y_lower+$grid_radius;
	$coor_k=($_[2]-$marginal_level)*$grid_side_length+$z_lower+$grid_radius;
	return($coor_i,$coor_j,$coor_k);
}


# extension of vdw radius for each atom
# &vdw_extension($i,\@ATOM,\@vdw_boundary)
sub vdw_extension
{
	my ($index, $pdb_aryref, $vdw_boundary_aryref)=@_;
	my ($i, $j, $k, $x, $y, $z, $atom_type, $level, $vdw_radius, $D_square, $Rr_square);
	my @box;

	$x=$$pdb_aryref[$index][14];
	$y=$$pdb_aryref[$index][15];
	$z=$$pdb_aryref[$index][16];
	$atom_type=$$pdb_aryref[$index][4];
	$vdw_radius=$Parameter::vdw_radius{$atom_type};
	$level=ceil($vdw_radius/$grid_side_length);
	$Rr_square=$vdw_radius**2;

	for( $i=$x-$level; $i<=$x+$level; $i++)
	{
		for( $j=$y-$level; $j<=$y+$level; $j++)
		{
			for( $k=$z-$level; $k<=$z+$level; $k++)
			{
				@box=();
				($box[0], $box[1], $box[2])=&inverse($i, $j, $k);
				$box[0]-=$$pdb_aryref[$index][9];
				$box[1]-=$$pdb_aryref[$index][10];
				$box[2]-=$$pdb_aryref[$index][11];
                                $D_square=$vector_op->norm_vec_square(@box);
                                if($D_square<=$Rr_square)
                                {
                                        $$vdw_boundary_aryref[$i][$j][$k][0]=1;
                                        push(@{$$vdw_boundary_aryref[$i][$j][$k][1]},$index);
                                }
                        }
                }
        }
}


### print a grid voxel in PDB format
sub print_pseduPDB
{
	my ($file_handle, $index, $resi, $coor_i, $coor_j, $coor_k)=@_;

	if($index<=99999)
	{
		printf $file_handle "ATOM  %5d  C   AAA   %3d    %8.3f%8.3f%8.3f\n",$index,$resi,$coor_i,$coor_j,$coor_k;
	}
	else
	{
		printf $file_handle "ATOM  *****  C   AAA   %3d    %8.3f%8.3f%8.3f\n",$resi,$coor_i,$coor_j,$coor_k;
	}
}


### scan PSP events along one dimension
### &perform_face_count('x',\@vdw_boundary,\@face_number,\@mouth_open);
sub perform_face_count
{
	my ($target_axe, $vdw_aryref, $face_number_aryref, $mouth_open_aryref)=@_;
	my (@terminal,@idx,@endi,@state);
	my ($x_idx,$y_idx,$z_idx,$mouth_idx,$first_meet,$second_meet,$total_meet);

	@terminal=($sol_x_endi,$sol_y_endi,$sol_z_endi);
	@idx=@endi=@state=();

	# define the loop variable index, 2 for inner most loop
	if($target_axe eq 'x')
	{
		$state[0]=2; $state[1]=1; $state[2]=0;
		$mouth_idx=0;
	}
	elsif($target_axe eq 'y')
	{
		$state[0]=0; $state[1]=2; $state[2]=1;
		$mouth_idx=1;
	}
	elsif($target_axe eq 'z')
	{
		$state[0]=1; $state[1]=0; $state[2]=2;
		$mouth_idx=2;
	}

	$endi[$state[0]]=$terminal[0];
	$endi[$state[1]]=$terminal[1];
	$endi[$state[2]]=$terminal[2];


	for($idx[0]=$sol_starti; $idx[0]<=$endi[0]; $idx[0]++)
	{
		for($idx[1]=$sol_starti; $idx[1]<=$endi[1]; $idx[1]++)
		{
			$first_meet=-1; $second_meet=-1; $total_meet=0;
			for($idx[2]=$sol_starti; $idx[2]<=$endi[2]; $idx[2]++)
			{
				$x_idx=$idx[$state[0]];
				$y_idx=$idx[$state[1]];
				$z_idx=$idx[$state[2]];
				if($$vdw_aryref[$x_idx][$y_idx][$z_idx][0]==1)
				{
					$first_meet=$idx[2];
					$total_meet++;
					last;
				}
			}

			if($first_meet<$endi[2] && $first_meet!=-1)
			{
				for($idx[2]=$endi[2];$idx[2]>$first_meet;$idx[2]--)
				{
					$x_idx=$idx[$state[0]];
					$y_idx=$idx[$state[1]];
					$z_idx=$idx[$state[2]];
					if($$vdw_aryref[$x_idx][$y_idx][$z_idx][0]==1)
					{
						$second_meet=$idx[2];
						$total_meet++;
						last;
					}
				}
			}

			if($total_meet==0)
			{
				next;
			}
			elsif($total_meet==1)
			{
				for($idx[2]=$sol_starti; $idx[2]<=$endi[2]; $idx[2]++)
				{
					$x_idx=$idx[$state[0]];
					$y_idx=$idx[$state[1]];
					$z_idx=$idx[$state[2]];
					$$face_number_aryref[$x_idx][$y_idx][$z_idx]+=2;
					$$mouth_open_aryref[$x_idx][$y_idx][$z_idx][$mouth_idx]=1;
				}
			}
			elsif($total_meet==2)
			{
				for($idx[2]=$sol_starti; $idx[2]<=$endi[2]; $idx[2]++)
				{
					$x_idx=$idx[$state[0]];
					$y_idx=$idx[$state[1]];
					$z_idx=$idx[$state[2]];
					if( $idx[2]<=$first_meet || $idx[2]>=$second_meet)
					{
						$$face_number_aryref[$x_idx][$y_idx][$z_idx]+=2;
						$$mouth_open_aryref[$x_idx][$y_idx][$z_idx][$mouth_idx]=1;
					}
					elsif( $idx[2]>$first_meet && $idx[2]<$second_meet)
					{
						$$face_number_aryref[$x_idx][$y_idx][$z_idx]+=20;
					}
				}
			}
		}
	}
}

### cylindrical scheme for scaning PSP events
sub orientational_complement
{
	my ($ATOM_ref,$vdw_boundary_ref,$face_number_ref)=@_;
	my (@pair_vector, @ray_trace, @v1_atom, @v2_atom, @v_a, @v_a_unit, @locate, @temp_scanned, @real_coor, @v_b, %perpendicular, %search_seq_hash);
	my ($m,$n,$o,$i,$j,$v_a_norm,$count_all,$D_bias,$r_1,$r_2,$locate_length);
	my ($target_length,$penertrate_flag,$grid_i,$grid_j,$grid_k,$ary_length);
	my ($r_cut,$r_cut_sq,$r_cut_level,$ab_dot,$v_b_norm_sq, $e_norm_sq, $ary_length,$idx_1,$idx_2,$cos_square, $grid_side_sq, $threshold);
	my ($user,$system,$child,$childsystem,$user1,$system1,$child1,$childsystem1);
	my  (@search_safe, %safe_hash, @search_safe_nonRD, @search_double_check, %double_check_hash, @search_double_check_nonRD);
	
	@pair_vector=@ray_trace=();
	%perpendicular=();
	$grid_side_sq=$grid_side_length**2;


	# check that vectors forming from each atom pairs don't penatrate through vdw expansion of protein atoms
	($user,$system,$child,$childsystem)=times;	
	$count_all=0;  $D_bias=3.4; # diameter of vdw radius of carbon
	for($i=0; $i<=$#$ATOM_ref-1; $i++)
	{
		@v1_atom=();
		@v1_atom=@{$$ATOM_ref[$i]}[9..11];
		
		for($j=$i+1; $j<=$#{$ATOM_ref}; $j++)
		{
			@v2_atom=();
			@v2_atom=@{$$ATOM_ref[$j]}[9..11];
		
			@v_a=(); $v_a_norm=0;
			$v_a[0]=$v2_atom[0]-$v1_atom[0];
			$v_a[1]=$v2_atom[1]-$v1_atom[1];
			$v_a[2]=$v2_atom[2]-$v1_atom[2];
		
			$v_a_norm=$vector_op->norm_vec(@v_a);
			if($v_a_norm > $D_bias)
			{
				@v_a_unit=();
				$v_a_unit[0]=$v_a[0]/$v_a_norm;
				$v_a_unit[1]=$v_a[1]/$v_a_norm;
				$v_a_unit[2]=$v_a[2]/$v_a_norm;

				$r_1=$Parameter::vdw_radius{$$ATOM_ref[$i][4]};
				$r_2=$Parameter::vdw_radius{$$ATOM_ref[$j][4]};
				$locate_length=$r_1+$grid_radius;
				$target_length=$v_a_norm-$r_2-$grid_radius;

				$penertrate_flag=0;
				while($locate_length<$target_length)
				{
					@locate=();
					$locate[0]=$v1_atom[0]+$locate_length*$v_a_unit[0];
					$locate[1]=$v1_atom[1]+$locate_length*$v_a_unit[1];
					$locate[2]=$v1_atom[2]+$locate_length*$v_a_unit[2];
					($grid_i,$grid_j,$grid_k)=&gridize(@locate);
					if($$vdw_boundary_ref[$grid_i][$grid_j][$grid_k][0]==1)
					{
						$penertrate_flag=1;
						last;
					}
					$locate_length+=$grid_side_length;
				}
				
				if($penertrate_flag==0)
				{
					$v_a[3]=$v_a_unit[0];
					$v_a[4]=$v_a_unit[1];
					$v_a[5]=$v_a_unit[2];
					$v_a[6]=$i;
					$v_a[7]=$j;
					$v_a[8]=$v_a_norm;
					@{$pair_vector[$count_all]}=@v_a;
					$count_all++;
				}
			}
		}
	}
	close FH;
	@v1_atom=(); @v2_atom=(); @v_a=(); $count_all=0;

	($user1,$system1,$child1,$childsystem1)=times;
	$user1-=$user;
	$system1-=$system;
	$child1-=$child;
	$childsystem1-=$childsystem;
	print "\n\ttotal $#pair_vector vectors\n";
	printf "\trunning time = %.1f (s)\n", $user1;

	($user,$system,$child,$childsystem)=times;
	%search_seq_hash=();
	#for($i=791; $i<=791; $i++)
	for($i=0; $i<=$#pair_vector; $i++)
	{
		$v_a_norm=$pair_vector[$i][8];
		@v_a_unit=();
		@v_a_unit=@{$pair_vector[$i]}[3..5];

		$r_1=$Parameter::vdw_radius{$$ATOM_ref[$pair_vector[$i][6]][4]};
		$r_2=$Parameter::vdw_radius{$$ATOM_ref[$pair_vector[$i][7]][4]};
		$r_cut=($r_1<=$r_2)?$r_1:$r_2;
		$r_cut_level=ceil($r_cut/$grid_side_length);
		$r_cut_sq=$r_cut**2;
		$r_safe=$r_cut-$grid_radius*sqrt(3);
		$threshold=($r_safe**2)/$grid_side_sq;  # square of r-(sqrt(3))*Side_length/2

		if( ! exists($search_seq_hash{$r_cut_level}))
		{
			&search_sequence($r_cut_level, \@{$search_seq_hash{$r_cut_level}});
		}

		### version 3 ###
		@search_safe=(); %safe_hash=(); @search_safe_nonRD=();
		@search_double_check=(); %double_check_hash=(); @search_double_check_nonRD=();
		for($j=0; $j<=$#{$search_seq_hash{$r_cut_level}}; $j++)
		{
			if($search_seq_hash{$r_cut_level}[$j][0] <= $threshold)
			{
				push(@search_safe,$search_seq_hash{$r_cut_level}[$j]);
				$safe_hash{$search_seq_hash{$r_cut_level}[$j][1]}{$search_seq_hash{$r_cut_level}[$j][2]}{$search_seq_hash{$r_cut_level}[$j][3]}=1;
			}
			else
			{
				push(@search_double_check,$search_seq_hash{$r_cut_level}[$j]);
				$double_check_hash{$search_seq_hash{$r_cut_level}[$j][1]}{$search_seq_hash{$r_cut_level}[$j][2]}{$search_seq_hash{$r_cut_level}[$j][3]}=1;
			}
		}
		
		@locate=();
		$locate[0]=$r_safe*$v_a_unit[0];
		$locate[1]=$r_safe*$v_a_unit[1];
		$locate[2]=$r_safe*$v_a_unit[2];
		($grid_i,$grid_j,$grid_k)=&gridize($locate[0],$locate[1],$locate[2]);
		for($j=0; $j<=$#search_safe; $j++)
		{
			$m=$grid_i+$search_safe[$j][1];
			$n=$grid_j+$search_safe[$j][2];
			$o=$grid_k+$search_safe[$j][3];

			if($safe_hash{$m}{$n}{$o} != 1)
			{
				push(@search_safe_nonRD,$search_safe[$j]);
			}
		}
		for($j=0; $j<=$#search_double_check; $j++)
		{
			$m=$grid_i+$search_double_check[$j][1];
			$n=$grid_j+$search_double_check[$j][2];
			$o=$grid_k+$search_double_check[$j][3];

			if($double_check_hash{$m}{$n}{$o} != 1)
			{
				push(@search_double_check_nonRD,$search_double_check[$j]);
			}
		}


		@v_a_atom=@{$$ATOM_ref[$pair_vector[$i][6]]}[9..11];
		@temp_scanned=();
		$locate_length=$r_1+$grid_radius;
		$target_length=$v_a_norm-$r_2-$grid_radius;
		$locate[0]=$v_a_atom[0]+$locate_length*$v_a_unit[0];
		$locate[1]=$v_a_atom[1]+$locate_length*$v_a_unit[1];
		$locate[2]=$v_a_atom[2]+$locate_length*$v_a_unit[2];
		($grid_i,$grid_j,$grid_k)=&gridize($locate[0],$locate[1],$locate[2]);
		for($j=0; $j<=$#search_safe; $j++)
		{
			$m=$grid_i+$search_safe[$j][1];
			$n=$grid_j+$search_safe[$j][2];
			$o=$grid_k+$search_safe[$j][3];

			$temp_scanned[$m][$n][$o]=1;
			if( $$vdw_boundary_ref[$m][$n][$o][0] != 1 && $$face_number_ref[$m][$n][$o]<40 )
			{
				push(@{$ray_trace[$m][$n][$o]},$i);
			}
		}
		$locate[0]=$v_a_atom[0]+$target_length*$v_a_unit[0];
		$locate[1]=$v_a_atom[1]+$target_length*$v_a_unit[1];
		$locate[2]=$v_a_atom[2]+$target_length*$v_a_unit[2];
		($grid_i,$grid_j,$grid_k)=&gridize($locate[0],$locate[1],$locate[2]);
		for($j=0; $j<=$#search_safe; $j++)
		{
			$m=$grid_i+$search_safe[$j][1];
			$n=$grid_j+$search_safe[$j][2];
			$o=$grid_k+$search_safe[$j][3];
			
			$temp_scanned[$m][$n][$o]=1;
			if( $$vdw_boundary_ref[$m][$n][$o][0] != 1 && $$face_number_ref[$m][$n][$o]<40 )
			{
				push(@{$ray_trace[$m][$n][$o]},$i);
			}
		}
		$locate_length+=$r_safe;
		
		while($locate_length<$target_length)
		{
			@locate=();
			$locate[0]=$v_a_atom[0]+$locate_length*$v_a_unit[0];
			$locate[1]=$v_a_atom[1]+$locate_length*$v_a_unit[1];
			$locate[2]=$v_a_atom[2]+$locate_length*$v_a_unit[2];
	
			($grid_i,$grid_j,$grid_k)=&gridize($locate[0],$locate[1],$locate[2]);
			for($j=0; $j<=$#search_safe_nonRD; $j++)
			{
				$m=$grid_i+$search_safe_nonRD[$j][1];
				$n=$grid_j+$search_safe_nonRD[$j][2];
				$o=$grid_k+$search_safe_nonRD[$j][3];
				if($temp_scanned[$m][$n][$o]!=1)
				{
					$temp_scanned[$m][$n][$o]=1;
					if( $$vdw_boundary_ref[$m][$n][$o][0] != 1 && $$face_number_ref[$m][$n][$o]<40 )
					{
						push(@{$ray_trace[$m][$n][$o]},$i);
					}
				}
			}
			$locate_length+=$r_safe;
		}
	

		$locate_length=$r_1+$grid_radius;
		$target_length=$v_a_norm-$r_2-$grid_radius;
		$locate[0]=$v_a_atom[0]+$locate_length*$v_a_unit[0];
		$locate[1]=$v_a_atom[1]+$locate_length*$v_a_unit[1];
		$locate[2]=$v_a_atom[2]+$locate_length*$v_a_unit[2];
		($grid_i,$grid_j,$grid_k)=&gridize($locate[0],$locate[1],$locate[2]);
		for($j=0; $j<=$#search_double_check; $j++)
		{
			$m=$grid_i+$search_double_check[$j][1];
			$n=$grid_j+$search_double_check[$j][2];
			$o=$grid_k+$search_double_check[$j][3];
			if($temp_scanned[$m][$n][$o]!=1)
			{

				$temp_scanned[$m][$n][$o]=1;
				if( $$vdw_boundary_ref[$m][$n][$o][0] != 1 && $$face_number_ref[$m][$n][$o]<40 )
				{
					@real_coor=();
					($real_coor[0],$real_coor[1],$real_coor[2])=&inverse($m,$n,$o);
					@v_b=();
					$v_b[0]=$real_coor[0]-$v_a_atom[0];
					$v_b[1]=$real_coor[1]-$v_a_atom[1];
					$v_b[2]=$real_coor[2]-$v_a_atom[2];
					$v_b_norm_sq=$vector_op->norm_vec_square(@v_b);
			
					$ab_dot=$vector_op->vec_dot(\@v_a_unit,\@v_b);
					$e_norm_sq=$v_b_norm_sq-$ab_dot**2;
					if($e_norm_sq<=$r_cut_sq)
					{
						push(@{$ray_trace[$m][$n][$o]},$i);
					}
				}
			}
		}
		$locate[0]=$v_a_atom[0]+$target_length*$v_a_unit[0];
		$locate[1]=$v_a_atom[1]+$target_length*$v_a_unit[1];
		$locate[2]=$v_a_atom[2]+$target_length*$v_a_unit[2];
		($grid_i,$grid_j,$grid_k)=&gridize($locate[0],$locate[1],$locate[2]);
		for($j=0; $j<=$#search_double_check; $j++)
		{
			$m=$grid_i+$search_double_check[$j][1];
			$n=$grid_j+$search_double_check[$j][2];
			$o=$grid_k+$search_double_check[$j][3];
		
			if($temp_scanned[$m][$n][$o]!=1)
			{
				$temp_scanned[$m][$n][$o]=1;
				if( $$vdw_boundary_ref[$m][$n][$o][0] != 1 && $$face_number_ref[$m][$n][$o]<40 )
				{
					@real_coor=();
					($real_coor[0],$real_coor[1],$real_coor[2])=&inverse($m,$n,$o);
					@v_b=();
					$v_b[0]=$real_coor[0]-$v_a_atom[0];
					$v_b[1]=$real_coor[1]-$v_a_atom[1];
					$v_b[2]=$real_coor[2]-$v_a_atom[2];
					$v_b_norm_sq=$vector_op->norm_vec_square(@v_b);
			
					$ab_dot=$vector_op->vec_dot(\@v_a_unit,\@v_b);
					$e_norm_sq=$v_b_norm_sq-$ab_dot**2;
					if($e_norm_sq<=$r_cut_sq)
					{
						push(@{$ray_trace[$m][$n][$o]},$i);
					}
				}
			}
		}
		$locate_length+=$r_safe;
		while($locate_length<$target_length)
		{
			@locate=();
			$locate[0]=$v_a_atom[0]+$locate_length*$v_a_unit[0];
			$locate[1]=$v_a_atom[1]+$locate_length*$v_a_unit[1];
			$locate[2]=$v_a_atom[2]+$locate_length*$v_a_unit[2];
	
			($grid_i,$grid_j,$grid_k)=&gridize($locate[0],$locate[1],$locate[2]);
			for($j=$#search_double_check_nonRD; $j>=0; $j--)
			{
				$m=$grid_i+$search_double_check_nonRD[$j][1];
				$n=$grid_j+$search_double_check_nonRD[$j][2];
				$o=$grid_k+$search_double_check_nonRD[$j][3];
				if($temp_scanned[$m][$n][$o]!=1)
				{

					$temp_scanned[$m][$n][$o]=1;
					if( $$vdw_boundary_ref[$m][$n][$o][0] != 1 && $$face_number_ref[$m][$n][$o]<40 )
					{
						@real_coor=();
						($real_coor[0],$real_coor[1],$real_coor[2])=&inverse($m,$n,$o);
						@v_b=();
						$v_b[0]=$real_coor[0]-$v_a_atom[0];
						$v_b[1]=$real_coor[1]-$v_a_atom[1];
						$v_b[2]=$real_coor[2]-$v_a_atom[2];
						$v_b_norm_sq=$vector_op->norm_vec_square(@v_b);
				
						$ab_dot=$vector_op->vec_dot(\@v_a_unit,\@v_b);
						$e_norm_sq=$v_b_norm_sq-$ab_dot**2;
						if($e_norm_sq<=$r_cut_sq)
						{
							push(@{$ray_trace[$m][$n][$o]},$i);
						}
					}
				}
			}
			$locate_length+=$r_safe;
		}
	}
	($user1,$system1,$child1,$childsystem1)=times;
	$user1-=$user;
	$system1-=$system;
	$child1-=$child;
	$childsystem1-=$childsystem;
	print "\n\tCylinder complete\n";
	printf "\trunning time = %.1f (s)\n", $user1;


	($user,$system,$child,$childsystem)=times;
	for($m=$sol_starti; $m<=$sol_x_endi; $m++)
	{
		for($n=$sol_starti; $n<=$sol_y_endi; $n++)
		{
			for($o=$sol_starti; $o<=$sol_z_endi; $o++)
			{
				$ary_length=$#{$ray_trace[$m][$n][$o]};
				if($ary_length>=1)
				{
				MARK_1:	for($i=0; $i<=$ary_length-1; $i++)
					{
						for($j=$i+1; $j<=$ary_length; $j++)
						{
							$idx_1=$ray_trace[$m][$n][$o][$i];
							$idx_2=$ray_trace[$m][$n][$o][$j];
							if( $perpendicular{$idx_1}{$idx_2}==1 )
							{
								$$face_number_ref[$m][$n][$o]=40;
								last MARK_1;
							}
							elsif( $perpendicular{$idx_1}{$idx_2}==2 )
							{
								next;
							}
							else
							{
								@v_a=(); @v_b=();
								@v_a=@{$pair_vector[$idx_1]}[0..2];
								@v_b=@{$pair_vector[$idx_2]}[0..2];
								$cos_square=$vector_op->cos_sq(@v_a,@v_b);
								if($cos_square<=0.007596124) # cos(85)**2 = cos(95)**2 = 0.007596124
								{
									$$face_number_ref[$m][$n][$o]=40;
									$perpendicular{$idx_1}{$idx_2}=1;
									$perpendicular{$idx_2}{$idx_1}=1;
									last MARK_1;
								}
								else
								{
									$perpendicular{$idx_1}{$idx_2}=2;
									$perpendicular{$idx_2}{$idx_1}=2;
								}
							}
						}
					}
				}
			}
		}
	}
	($user1,$system1,$child1,$childsystem1)=times;
	$user1-=$user;
	$system1-=$system;
	$child1-=$child;
	$childsystem1-=$childsystem;
	print "\n\tCheck perpendicularity complete\n";
	printf "\trunning time = %.1f (s)\n", $user1;

	#free memory
	%perpendicular=();
	%ray_trace=();
}

### cluster pore grid voxels to form disjoint pores
### &cluster(\@void_nucleus,\@pore_cluster,\@pore_volume,\@search_seq_1);
sub cluster
{
	my ($I,$J,$K,$total); my $cluster_index=0;
	my ($void_nucleus_aryref,$pore_cluster_aryref,$pore_volume_aryref, $search_seq_1_aryref)=@_;

	for($I=$pro_starti;$I<=$pro_x_endi;$I++)
	{
		for($J=$pro_starti;$J<=$pro_y_endi;$J++)
		{
			for($K=$pro_starti;$K<=$pro_z_endi;$K++)
			{
				if( $$void_nucleus_aryref[$I][$J][$K]==1 && $$pore_cluster_aryref[$I][$J][$K]<1 )
				{
					$total=0;
					$cluster_index++;
					$total=&neighbor_join($I, $J, $K, $void_nucleus_aryref, $pore_cluster_aryref, $pore_volume_aryref, $cluster_index, $search_seq_1_aryref);
					$$pore_volume_aryref[$cluster_index][0]=$cluster_index;
					$$pore_volume_aryref[$cluster_index][1]=$total;
				}
			}
		}
	}
}

### main subroutin of cluster
sub neighbor_join
{
	my ($x, $y, $z, $void_nucleus_aryref, $pore_cluster_aryref, $pore_volume_aryref, $index, $search_seq_1_aryref )=@_;
	my ($sum, $idx_temp, $i, $j, $k, $m);


	$$pore_cluster_aryref[$x][$y][$z]=$index;

	$$pore_volume_aryref[$index][0]=$index;
	$idx_temp=$#{$$pore_volume_aryref[$index][2]};
	$idx_temp++;
	$$pore_volume_aryref[$index][2][$idx_temp][0]=$x;
	$$pore_volume_aryref[$index][2][$idx_temp][1]=$y;
	$$pore_volume_aryref[$index][2][$idx_temp][2]=$z;

	$sum=1;
	for($m=1; $m<=$#{$search_seq_1_aryref}; $m++)
	{
		$i=$x+$$search_seq_1_aryref[$m][1];
		$j=$y+$$search_seq_1_aryref[$m][2];
		$k=$z+$$search_seq_1_aryref[$m][3];
		if( $$void_nucleus_aryref[$i][$j][$k]==1 && $$pore_cluster_aryref[$i][$j][$k]<1 )
		{
			$sum+=&neighbor_join($i, $j, $k, $void_nucleus_aryref, $pore_cluster_aryref, $pore_volume_aryref, $index, $search_seq_1_aryref);
		}
	}
	return($sum);
}


### complete the volume of the pore according to the size of the probe
### &complete_volume(\@pore_volume, \@void_nucleus, \@face_number, \@pore_cluster, \@search_seq_probe, \@search_seq_1);
sub complete_volume
{
	my ($pore_volume_aryref, $void_nucleus_aryref, $face_number_aryref, $pore_cluster_aryref, $search_seq_probe_aryref, $search_seq_1_aryref)=@_; 
	my ($length1, $length2, $appendN, $D_sq, $grid_side_sq, $i, $j, $k, $l, $p, $q, $r, $m, $n, $o);
	my (@tempx,@tempy,@tempz,@tempC);
	my $level=ceil(($solvent_radius-$grid_radius)/$grid_side_length);
	my $sol_Rr_sq=$solvent_radius**2;
	my $grid_side_sq=$grid_side_length**2;	

	$length1=$#{$pore_volume_aryref};
	for( $i=0; $i<$length1; $i++ )
	{
		$appendN=0;
		@tempx=@tempy=@tempz=@tempC=();
		$length2=$#{$$pore_volume_aryref[$i][2]};
		for( $j=0;$j<=$length2; $j++ )
		{
			$x=$$pore_volume_aryref[$i][2][$j][0];
			$y=$$pore_volume_aryref[$i][2][$j][1];
			$z=$$pore_volume_aryref[$i][2][$j][2];
			for($k=1; $k<=$#{$search_seq_probe_aryref}; $k++)
			{
				$p=$x+$$search_seq_probe_aryref[$k][1];
				$q=$y+$$search_seq_probe_aryref[$k][2];
				$r=$z+$$search_seq_probe_aryref[$k][3];
				if( $tempC[$p][$q][$r]!=1 && $$void_nucleus_aryref[$p][$q][$r]!=1 && $$face_number_aryref[$p][$q][$r] >=40)
				{
					$D_sq=$$search_seq_probe_aryref[$k][0]*$grid_side_sq;
					if( $D_sq <= $sol_Rr_sq )
					{
						for($l=1; $l<=$#{$search_seq_1_aryref}; $l++)
						{
							$m=$p+$$search_seq_1_aryref[$l][1];
							$n=$q+$$search_seq_1_aryref[$l][2];
							$o=$r+$$search_seq_1_aryref[$l][3];
							if($$pore_cluster_aryref[$m][$n][$o]==$$pore_volume_aryref[$i][0])
							{
								$tempC[$p][$q][$r]=1;
								$$pore_cluster_aryref[$p][$q][$r]=$$pore_volume_aryref[$i][0];
								$appendN++;
								push(@tempx,$p); push(@tempy,$q); push(@tempz,$r);
								last;
							}
						}
					}
				}
			}
		}

		$$pore_volume_aryref[$i][1]+=$appendN;
		for( $j=1; $j<=$appendN; $j++ )
		{
			$$pore_volume_aryref[$i][2][$length2+$j][0]=$tempx[$j-1];
			$$pore_volume_aryref[$i][2][$length2+$j][1]=$tempy[$j-1];
			$$pore_volume_aryref[$i][2][$length2+$j][2]=$tempz[$j-1];
		}
	}
}


### the purpose of pore_surface is to define the surface of the pore region that can help defining pore axis
### &pore_surface(\@pore_volume, \@pore_cluster, \@face_number, \@vdw_boundary, \@COM);
sub pore_surface
{
	my ($pore_volume_aryref, $pore_cluster_aryref, $face_number_aryref, $vdw_boundary_aryref, $COM_aryref )=@_;
	my ($iter_1, $iter_2, $i, $j, $k, $x, $y, $z, $sum, $touch_sum);
	my @vec_tmp;

	$iter_1=$#{$pore_volume_aryref};
	for($i=0; $i<$iter_1; $i++)
	{
		$iter_2=$#{$$pore_volume_aryref[$i][2]};
		for($j=0; $j<=$iter_2; $j++)
		{
			@vec_tmp=();
			for($k=0; $k<=2; $k++)
			{
				$vec_tmp[$k]=$COM[$k]-$$pore_volume_aryref[$i][2][$j][$k];
			}
			$$pore_volume_aryref[$i][2][$j][4]=$vector_op->norm_vec(@vec_tmp);
			
			
			$sum=0; $touch_sum=0;
			($x, $y, $z)=@{$$pore_volume_aryref[$i][2][$j]};
			
			if($$pore_cluster_aryref[$x+1][$y][$z]==$$pore_volume_aryref[$i][0])
			{
				$sum++;
			}
			elsif($$face_number_aryref[$x+1][$y][$z]<40 && $$vdw_boundary_aryref[$x+1][$y][$z][0]!=1)
			{
				$touch_sum++;
			}
			
			if($$pore_cluster_aryref[$x-1][$y][$z]==$$pore_volume_aryref[$i][0])
			{
				$sum++;
			}
			elsif($$face_number_aryref[$x-1][$y][$z]<40 && $$vdw_boundary_aryref[$x-1][$y][$z][0]!=1)
			{
				$touch_sum++;
			}
			
			if($$pore_cluster_aryref[$x][$y+1][$z]==$$pore_volume_aryref[$i][0])
			{
				$sum++;
			}
			elsif($$face_number_aryref[$x][$y+1][$z]<40 && $$vdw_boundary_aryref[$x][$y+1][$z][0]!=1)
			{
				$touch_sum++;
			}
			
			if($$pore_cluster_aryref[$x][$y-1][$z]==$$pore_volume_aryref[$i][0])
			{
				$sum++;
			}
			elsif($$face_number_aryref[$x][$y-1][$z]<40 && $$vdw_boundary_aryref[$x][$y-1][$z][0]!=1)
			{
				$touch_sum++;
			}
			
			if($$pore_cluster_aryref[$x][$y][$z+1]==$$pore_volume_aryref[$i][0])
			{
				$sum++;
			}
			elsif($$face_number_aryref[$x][$y][$z+1]<40 && $$vdw_boundary_aryref[$x][$y][$z+1][0]!=1)
			{
				$touch_sum++;
			}
			
			if($$pore_cluster_aryref[$x][$y][$z-1]==$$pore_volume_aryref[$i][0])
			{
				$sum++;
			}
			elsif($$face_number_aryref[$x][$y][$z-1]<40 && $$vdw_boundary_aryref[$x][$y][$z-1][0]!=1)
			{
				$touch_sum++;
			}

			if($sum<6)
			{
				if( $sum+$touch_sum == 6 )
				{
					$$pore_volume_aryref[$i][2][$j][3]=2; # 2 for pore surface outside the protein
				}
				else
				{
					$$pore_volume_aryref[$i][2][$j][3]=1; # 1 for pore surface inside the protein
				}
			}
			else
			{
				$$pore_volume_aryref[$i][2][$j][3]=0;
			}
		}
	}
}

### An effecient search sequence for scanning a specified region in the grid
### &search_sequence($level, \@search_seq);
sub search_sequence
{
	my ($level,$search_seq_aryref)=@_;
	my ($count, $i, $j, $k);
	my @vec;

	$count=0;
	for($i=-$level; $i<=$level; $i++)
	{
		for($j=-$level; $j<=$level; $j++)
		{
			for($k=-$level; $k<=$level; $k++)
			{
				@vec=();
				@vec=($i,$j,$k);
				$$search_seq_aryref[$count][0]=$vector_op->norm_vec_square(@vec);
				$$search_seq_aryref[$count][1]=$i;
				$$search_seq_aryref[$count][2]=$j;
				$$search_seq_aryref[$count][3]=$k;
				$count++;
			}
		}
	}

	@{$search_seq_aryref}=sort{$a->[0] <=> $b->[0]} (@{$search_seq_aryref});
}


### define lining residue of an identified pore
### &extract_lining_residue(\@pore_volume,$pore_rank,\@vdw_boundary,\%lining_residue);
sub extract_lining_residue
{
	my $array_ref1=$_[0]; my $array_ref2=$_[2]; my $hash_ref=$_[3];
	my $index=$_[1];
	my ( $length, $length2, $index2, $x, $y, $z, $i, $l, $m, $n, $p);


	$length=$#{$$array_ref1[$index][2]};
	for( $i=0; $i<=$length; $i++ )
	{
		$x=$$array_ref1[$index][2][$i][0];
		$y=$$array_ref1[$index][2][$i][1];
		$z=$$array_ref1[$index][2][$i][2];
		for( $l=-1+$x; $l<=1+$x; $l++ )
		{
			for( $m=-1+$y; $m<=1+$y; $m++ )
			{
				for( $n=-1+$z; $n<=1+$z; $n++ )
				{
					if($$array_ref2[$l][$m][$n][0]==1)
					{
						$length2=$#{$$array_ref2[$l][$m][$n][1]};
						for( $p=0; $p<=$length2; $p++ )
						{
							$index2=$$array_ref2[$l][$m][$n][1][$p];
							if($$hash_ref{$ATOM[$index2][7]}{$ATOM[$index2][8]} eq "" )
							{
								$$hash_ref{$ATOM[$index2][7]}{$ATOM[$index2][8]}=$ATOM[$index2][6];
							}
						}
					}
				}
			}
		}
	}
}


### unite the void grid voxel outside the protein to seal the gap of pores in side the protein
sub union
{
	my ($x, $y, $z, $background_aryref, $not_covered_raw_aryref, $search_seq_1_aryref)=@_;
	my ($i, $p, $q, $r);
	
	if($x>=$sol_starti && $x<=$sol_x_endi && $y>=$sol_starti && $y<=$sol_y_endi && $z>=$sol_starti && $z<=$sol_z_endi)
	{
		for($i=1; $i<=$#$search_seq_1_aryref; $i++)
		{
			$p=$x+$$search_seq_1_aryref[$i][1];
			$q=$y+$$search_seq_1_aryref[$i][2];
			$r=$z+$$search_seq_1_aryref[$i][3];
			if( $$not_covered_raw_aryref[$p][$q][$r]==1 && $$background_aryref[$p][$q][$r]!=1)
			{
				$$background_aryref[$p][$q][$r]=1;
				&union($p, $q, $r, $background_aryref, $not_covered_raw_aryref, $search_seq_1_aryref);
			}
		}
	}
}


### get and check the input parameters
sub get_check_options
{
	my %opts=();
	my ($pdb_file, $grid_side_length, $probe_radius, $shallow_criteria, $output_prefix)=("", 1.0, 1.2, 1.4, ""); # defaults values

	getopt('frscnh',\%opts);

	if( exists $opts{'h'} )
	{
		&help_test();
		exit;
	}
	
	if( !exists $opts{'f'} )
	{
		print "\nPlease input pdb file of protein\n";
		&help_test();
		exit;
	}
	else
	{
		$pdb_file=$opts{'f'};
	}

	if( exists $opts{'r'} )
	{
		$grid_side_length=$opts{'r'};
	}

	if( exists $opts{'s'} )
	{
		$probe_radius=$opts{'s'};
	}
	
	if( exists $opts{'c'} )
	{
		$shallow_criteria=$opts{'c'};
	}

	if( exists $opts{'n'} )
	{
		$output_prefix=$opts{'n'};	
	}
	else
	{
		if($pdb_file=~/(\S+)\.pdb/ || $pdb_file=~/(\S+)\.PDB/)
		{
			$output_prefix=$1;	
		}
	}


	### check for some mistakes ###
	if( $pdb_file eq "")
	{
		print "\nPlease input pdb file of protein\n";
		&help_test();
		exit;
	}

	if($grid_side_length < 0 || $probe_radius<0 || $shallow_criteria<0)
	{
		print "\nAll parameters must be positive values.\n";
		&help_test();
		exit;
	}

	if( $grid_side_length>=$probe_radius || $grid_side_length>=$shallow_criteria )
	{
		print "\nThe 'side length of the grid voxel' must be smaller than either the 'probe radius' or the 'criteria for shallow pore region'.\n";
		&help_test();
		exit;
	}

	print "\nRunning perl Pore_ID.pl -f $pdb_file -r $grid_side_length -s $probe_radius -c $shallow_criteria -n $output_prefix\n\n";
	return($pdb_file, $grid_side_length, $probe_radius, $shallow_criteria, $output_prefix);
}


### usage of this program
sub help_test
{
	print "\nusage : perl Pore_ID.pl -f pdb_file [-r 1.0 -s 1.2 -c 1.4 -n]\n";
	print "\t-f\tPDB file of the targeted protein\n";
	print "\t-r\tside length of the grid voxel (default: 1.0 angstrom)\n";
	print "\t-s\tprobe radius (default: 1.2 angstrom)\n";
	
	print "\t-c\ttrimming depth for shallow pore region on the protein surface (default: 1.4 angstrom)\n";
	print "\t-n\tprefix of output files (default: the prefix of the .pdb file)\n";
	print "\t-h\tshow this page\n\n";
}
