############################################################################
#                                                                          #
#   Copyright (C) 2011 Po-Hsien Lee <p.lee@mx.uni-saarland.de>             #
#   All rights reserved.                                                   #
#                                                                          #
#   This file is part of PROPORES.                                         #
#                                                                          #
#   PROPORES is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation, either version 3 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   PROPORES is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License      #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
#                                                                          #
############################################################################
#!/usr/bin/perl


package KD_tree_NN;

use Parameter;
use POSIX;


sub new
{
	my $type = $_[0];
	my $self = {};
	bless($self, $type);
	return($self);
}


### construct a KD tree of a protein structure (PDB file)
sub construct_kd_tree 
{
	my $class = shift @_;
	my($ATOM_ref, $depth) = @_;
	my($axe_idx,$median,$med_idx, $tmp_med);
	my(@sorted_ATOM, @first_half);
	my $tree={};

	if($#$ATOM_ref==0)
	{
		$tree->{VALUE}=$$ATOM_ref[0][17];
	}
	elsif($#$ATOM_ref<0)
	{
		$tree->{VALUE}=-99999;
	}
	else 
	{
		$axe_idx=($depth%3)+9;
		@sorted_ATOM=sort{ $a->[$axe_idx] <=> $b->[$axe_idx] }(@{$ATOM_ref});
		$med_idx=int($#sorted_ATOM/2);
		$tmp_med=$sorted_ATOM[$med_idx][$axe_idx];
		while($tmp_med==$sorted_ATOM[$med_idx+1][$axe_idx])
		{
			$med_idx++;
		}
		$median=0.5*($tmp_med+$sorted_ATOM[$med_idx+1][$axe_idx]);
		@first_half=splice(@sorted_ATOM,0,$med_idx+1);
		$tree->{VALUE}=$median;
		$tree->{LEFT}=KD_tree_NN->construct_kd_tree(\@first_half,$depth+1);
		$tree->{RIGHT}=KD_tree_NN->construct_kd_tree(\@sorted_ATOM,$depth+1);
	}
	return $tree;
}


### initial run of KD search 
sub nnsearch_kd_tree_r1
{
	my $class = shift @_;
	my ($tree, $depth, $ATOM_ref, @value) = @_;
	my ($axe_idx, $node, $atom_idx, $radius, $i);
	my (@tmp_vec);

	$node=$tree;
	$axe_idx=($depth%3);
	
	if( defined($node->{LEFT}) )
	{
		if($value[$axe_idx]<$node->{VALUE})
		{
			($atom_idx,$radius)=KD_tree_NN->nnsearch_kd_tree_r1($node->{LEFT},$depth+1,$ATOM_ref,@value);
		}
		else
		{
			($atom_idx,$radius)=KD_tree_NN->nnsearch_kd_tree_r1($node->{RIGHT},$depth+1,$ATOM_ref,@value);
		}
	}
	else
	{
		$atom_idx=$node->{VALUE};
		if($atom_idx==-99999) # locate in an empty hyperrectangle
		{
			$radius=99999;
		}
		else
		{
			$radius=0;
			@tmp_vec=();
			@tmp_vec=@{$$ATOM_ref[$atom_idx]}[9..11];
			for($i=0; $i<=2; $i++)
			{
				$tmp_vec[$i]-=$value[$i];
				$radius+=$tmp_vec[$i]**2;
			}
			$radius=sqrt($radius)-$Parameter::vdw_radius{$$ATOM_ref[$atom_idx][4]};
		}
	}
	return ($atom_idx,$radius);
}


### find the largest ball that can fit into the void grid (vdw boundary) after the initial Kd search by nnsearch_kd_tree_r1
sub nnsearch_kd_tree_r2
{
	my $class = shift @_;
	my ($tree, $depth, $idx_ref, $min_ref, $ATOM_ref, @value)=@_;
	my ($node, $axe_idx, $distance, $atom_idx, $radius, $S_vdw_r);
	my (@tmp_vec);

	$node=$tree;
	$axe_idx=$depth%3;
	$S_vdw_r=$Parameter::vdw_radius{'S'};

	if( defined($node->{LEFT}) )
	{
		$distance=$node->{VALUE}-$value[$axe_idx];
		
		if(abs($distance) < $$min_ref+$S_vdw_r)
		{
			KD_tree_NN->nnsearch_kd_tree_r2($node->{LEFT},$depth+1,$idx_ref, $min_ref, $ATOM_ref, @value);
			KD_tree_NN->nnsearch_kd_tree_r2($node->{RIGHT},$depth+1,$idx_ref, $min_ref, $ATOM_ref, @value);
		}
		else
		{
			if($distance<0)
			{
				KD_tree_NN->nnsearch_kd_tree_r2($node->{RIGHT},$depth+1,$idx_ref, $min_ref, $ATOM_ref, @value);
			}
			else
			{
				KD_tree_NN->nnsearch_kd_tree_r2($node->{LEFT},$depth+1,$idx_ref, $min_ref, $ATOM_ref, @value);
			}
		}
	}
	else
	{
		$atom_idx=$node->{VALUE};
		
		if($atom_idx==-99999) # locate in an empty hyperrectangle
		{
			$radius=99999;
		}
		else
		{
			$radius=0;
			@tmp_vec=();
			@tmp_vec=@{$$ATOM_ref[$atom_idx]}[9..11];
			for($i=0; $i<=2; $i++)
			{
				$tmp_vec[$i]-=$value[$i];
				$radius+=$tmp_vec[$i]**2;
			}
			$radius=sqrt($radius)-$Parameter::vdw_radius{$$ATOM_ref[$atom_idx][4]};
		}

		if( $radius < $$min_ref )
		{
			$$min_ref=$radius;
			$$idx_ref=$atom_idx;
		}
		return 1;
	}
}


return(1);
